/*
 * Copyright 2006 Robert Burrell Donkin
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */ 
package rat.element;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Locale;

public class BinaryElement extends AbstractElement {

    private static final int HIGH_BYTES_RATIO = 100;
    private static final int TOTAL_READ_RATIO = 30;
    private static final int NON_ASCII_THREASHOLD = 256;
    private static final int ASCII_CHAR_THREASHOLD = 8;

    private static final String JAR_MANIFEST = "MANIFEST.MF";
    private static final String JAVA = "JAVA";

    /**
     * Based on http://www.apache.org/dev/svn-eol-style.txt
     */
    private static final String[] NON_BINARY_EXTENSIONS = {
        "AART",
        "AC",
        "AM",
        "BAT",
        "C",
        "CAT",
        "CGI",
        "CLASSPATH",
        "CMD",
        "CONFIG",
        "CPP",
        "CSS",
        "CWIKI",
        "DATA",
        "DCL",
        "DTD",
        "EGRM",
        "ENT",
        "FT", 
        "FN",
        "FV", 
        "GRM",
        "G",
        "H",
        "HTACCESS",
        "HTML",
        "IHTML",
        "IN",
        "JAVA",
        "JMX", 
        "JSP",
        "JS",
        "JUNIT",
        "JX", 
        "MANIFEST",
        "M4",
        "MF",
        "MF",
        "META",
        "MOD",
        "N3",
        "PEN",
        "PL",
        "PM",
        "POD",
        "POM",
        "PROJECT",
        "PROPERTIES",
        "PY",
        "RB",
        "RDF",
        "RNC",
        "RNG",
        "RNX",
        "ROLES",
        "RSS",
        "SH",
        "SQL",
        "SVG",
        "TLD",
        "TXT",
        "TYPES",
        "VM",
        "VSL",
        "WSDD",
        "WSDL",
        "XARGS",
        "XCAT",
        "XCONF",
        "XEGRM",
        "XGRM",
        "XLEX",
        "XLOG",
        "XMAP",
        "XML",
        "XROLES",
        "XSAMPLES",
        "XSD",
        "XSL",
        "XSLT",
        "XSP",
        "XUL",
        "XWEB",
        "XWELCOME",
    };

	public BinaryElement(File file) {
		super(file);
	}

    public BinaryElement(String name) {
        super(name, null);
    }

	public void accept(IElementVisitor visitor) {
		visitor.visit(this);
	}
	
	
	public String toString() {
		return "  B     " + getName();
	}

    /**
     * Is a file by that name a known binary file?
     */
    public static final boolean isBinary(final String name) {
        if (name == null) {return false;}
        String nameToUpper = name.toUpperCase(Locale.US);
        return JAR_MANIFEST.equals(name) || isImage(nameToUpper)
            || isKeystore(nameToUpper) || isBytecode(nameToUpper)
            || isBinaryData(nameToUpper) || isExecutable(nameToUpper);
    }

    /**
     * Is a file by that name a known non-binary file?
     */
    public static final boolean isNonBinary(final String name) {
        if (name == null) {return false;}
        return extensionMatches(name.toUpperCase(Locale.US),
                                NON_BINARY_EXTENSIONS);
    }

    private static final String[] DATA_EXTENSIONS = {
        "DAT", "DOC",
        "NCB", "IDB",
        "SUO", "XCF",
        "RAJ", "CERT",
        "KS", "TS",
    };

    static final boolean isBinaryData(final String name) {
        return extensionMatches(name, DATA_EXTENSIONS);
    }
    
    private static final String[] EXE_EXTENSIONS = {
        "EXE", "DLL",
        "LIB", "SO",
        "A", "EXP",
    };

    static final boolean isExecutable(final String name) {
        return name.equals(JAVA) || extensionMatches(name, EXE_EXTENSIONS)
            || containsExtension(name, EXE_EXTENSIONS);
    }

    private static final String[] KEYSTORE_EXTENSIONS = {
        "JKS", "KEYSTORE",
    };

    static final boolean isKeystore(final String name) {
        return extensionMatches(name, KEYSTORE_EXTENSIONS);
    }
	
    private static final String[] IMAGE_EXTENSIONS = {
        "PNG", "PDF",
        "GIF", "GIFF",
        "TIF", "TIFF",
        "JPG", "JPEG",
        "ICO",
    };

    static final boolean isImage(final String name) {
        return extensionMatches(name, IMAGE_EXTENSIONS);
    }

    private static final String[] BYTECODE_EXTENSIONS = {
        "CLASS", "PYD",
        "OBJ", "PYC",
    };

    static boolean isBytecode(final String name) {
        return extensionMatches(name, BYTECODE_EXTENSIONS);
    }

    private static boolean extensionMatches(final String name,
                                            final String[] exts) {
        boolean result = false;
        for (int i = 0; !result && i < exts.length; i++) {
            result = name.endsWith("." + exts[i]);
        }
        return result;
    }

    private static boolean containsExtension(final String name,
                                             final String[] exts) {
        boolean result = false;
        for (int i = 0; !result && i < exts.length; i++) {
            result = name.indexOf("." + exts[i] + ".") >= 0;
        }
        return result;
    }

    /**
     * Do the first few bytes of the stream hint at a binary file?
     */
    public static boolean isBinary(InputStream in) {
        boolean result = false;
        byte[] taste = new byte[100];
        try {
            int bytesRead = in.read(taste);
            if (bytesRead > 0) {
                int highBytes = 0;
                for (int i=0;i<bytesRead;i++) {
                    if (taste[i] > NON_ASCII_THREASHOLD
                        || taste[i] <= ASCII_CHAR_THREASHOLD) {
                        highBytes++;
                    }
                }
                if (highBytes * HIGH_BYTES_RATIO
                    > bytesRead * TOTAL_READ_RATIO) {
                    result = true;
                }
            }
        } catch (IOException e) {
            // SWALLOW 
        }
        return result;
    }

}
